#include <windows.h>

#define OHBEXPORT __declspec(dllexport)
#include "ohbbhlp.h"
#include "ohbabel.h"

#define OHBINITLOCK() InitializeCriticalSection(&CriticalSection)
#define OHBLOCK() EnterCriticalSection(&CriticalSection)
#define OHBUNLOCK() LeaveCriticalSection(&CriticalSection)
#define OHBDESTROYLOCK() DeleteCriticalSection(&CriticalSection)

#define OHB_TRANS_WIN_CLS "TransWindow"
#define OHB_HOME_DIR_LEN 2048
#define OHB_JK_TRANS_MSG 5393
#define OHB_KJ_TRANS_MSG 5409

typedef struct _OHBABEL
{
  CHAR szJKHomeDir[OHB_HOME_DIR_LEN];
  CHAR szKJHomeDir[OHB_HOME_DIR_LEN];
  HMODULE hJKMod;
  HMODULE hKJMod;
  int (*InitJKTE)(HWND hwnd, const char *homeDir, int data0);
  int (*SetJKTESource)(int fd, const char *srcStr);
  int (*SetJKTEOption)(int fd, int p1, int p2, int p3, int p4, int p5,
    int p6, int p7, int p8, int p9, int p10, int p11, int p12, int p13,
    int p14, const char *srcStr);
  int (*TransJKTE)(int fd, int data0);
  const char *(*GetJKTESolution)(int fd);
  int (*ReleaseJKTESolution)(int fd);
  int (*EndJKTE)(int fd, int fd2);
  int (*InitKJTE)(HWND hwnd, const char *homeDir, int data0);
  int (*SetKJTESource)(int fd, const char *srcStr);
  int (*SetKJTEOption)(int fd, int p1, int p2, int p3, int p4, int p5,
    const char *srcStr);
  int (*TransKJTE)(int fd, int data0);
  const char *(*GetKJTESolution)(int fd);
  int (*ReleaseKJTESolution)(int fd);
  int (*EndKJTE)(int fd, int fd2);
} OHBABEL;

static OHBABEL babel;
static CRITICAL_SECTION CriticalSection;
static int bInit;

static int OhbBabelInit(void);
static int OhbBabelDestroy(void);
static HWND OhbCreateTransWindow(void);
static int OhbDestroyTransWindow(HWND hwndTransWindow);
static int OhbInitHomeDir(void);
static int OhbDestroyHomeDir(void);
static int OhbInitBabel(void);
static int OhbDestroyBabel(void);
static LRESULT CALLBACK OhbTransWindowMsgProc(HWND hwnd, UINT message,
  WPARAM wParam, LPARAM lParam);

static int OhbBabelInit(void)
{
  OHBINITLOCK();
  if (OhbInitHomeDir() != 0)
  {
    return -1;
  }
  if (OhbInitBabel() != 0)
  {
    return -2;
  }

  return 0;
}

static int OhbBabelDestroy(void)
{
  OhbDestroyBabel();
  OhbDestroyHomeDir();
  OHBDESTROYLOCK();

  return 0;
}


static HWND OhbCreateTransWindow(void)
{
  WNDCLASS wndclass;

  wndclass.style = 0;
  wndclass.lpfnWndProc = OhbTransWindowMsgProc;
  wndclass.cbClsExtra = 0;
  wndclass.cbWndExtra = 0;
  wndclass.hInstance = 0;
  wndclass.hIcon = 0;
  wndclass.hCursor = 0;
  wndclass.hbrBackground = 0;
  wndclass.lpszMenuName = 0;
  wndclass.lpszClassName = TEXT(OHB_TRANS_WIN_CLS);

  RegisterClass(&wndclass);

  return CreateWindow(TEXT(OHB_TRANS_WIN_CLS), TEXT("TransApp"), 0, 0, 0, 0,
    0, 0, 0, 0, 0);
}

static int OhbDestroyTransWindow(HWND hwndTransWindow)
{
  DestroyWindow(hwndTransWindow);
  UnregisterClass(TEXT(OHB_TRANS_WIN_CLS), 0);
  return 0;
}

static int OhbInitBabel(void)
{
  CHAR szDllBuff[OHB_HOME_DIR_LEN];

  lstrcpyA(szDllBuff, babel.szJKHomeDir);
  lstrcatA(szDllBuff, "..\\jkte40.dll");
  babel.hJKMod = LoadLibraryA(szDllBuff);
  if (babel.hJKMod == NULL)
  {
    return -1;
  }

  lstrcpyA(szDllBuff, babel.szKJHomeDir);
  lstrcatA(szDllBuff, "..\\kjtedll.dll");
  babel.hKJMod = LoadLibraryA(szDllBuff);
  if (babel.hKJMod == NULL)
  {
    return -2;
  }

  babel.InitJKTE = (LPVOID)GetProcAddress(babel.hJKMod, "InitJKTE");
  babel.SetJKTESource = (LPVOID)GetProcAddress(babel.hJKMod, "SetJKTESource");
  babel.SetJKTEOption = (LPVOID)GetProcAddress(babel.hJKMod, "SetJKTEOption");
  babel.TransJKTE= (LPVOID)GetProcAddress(babel.hJKMod, "TransJKTE");
  babel.GetJKTESolution = (LPVOID)GetProcAddress(babel.hJKMod,
    "GetJKTESolution");
  babel.ReleaseJKTESolution = (LPVOID)GetProcAddress(babel.hJKMod,
    "ReleaseJKTESolution");
  babel.EndJKTE= (LPVOID)GetProcAddress(babel.hJKMod, "EndJKTE");

  babel.InitKJTE = (LPVOID)GetProcAddress(babel.hKJMod, "InitKJTE");
  babel.SetKJTESource = (LPVOID)GetProcAddress(babel.hKJMod, "SetKJTESource");
  babel.SetKJTEOption = (LPVOID)GetProcAddress(babel.hKJMod, "SetKJTEOption");
  babel.TransKJTE= (LPVOID)GetProcAddress(babel.hKJMod, "TransKJTE");
  babel.GetKJTESolution = (LPVOID)GetProcAddress(babel.hKJMod,
    "GetKJTESolution");
  babel.ReleaseKJTESolution = (LPVOID)GetProcAddress(babel.hKJMod,
    "ReleaseKJTESolution");
  babel.EndKJTE= (LPVOID)GetProcAddress(babel.hKJMod, "EndKJTE");

  if (babel.InitJKTE == NULL || babel.SetJKTESource == NULL ||
    babel.SetJKTEOption == NULL || babel.TransJKTE == NULL ||
    babel.GetJKTESolution == NULL || babel.ReleaseJKTESolution == NULL ||
    babel.EndJKTE == NULL || babel.InitKJTE == NULL ||
    babel.SetKJTESource == NULL || babel.SetKJTEOption == NULL ||
    babel.TransKJTE == NULL || babel.GetKJTESolution == NULL ||
    babel.ReleaseKJTESolution == NULL || babel.EndKJTE == NULL)
  {
    return -3;
  }

  return 0;
}

static int OhbDestroyBabel(void)
{
  if (babel.hJKMod)
  {
/*    FreeLibrary(babel.hJKMod); */
  }

  if (babel.hKJMod)
  {
/*    FreeLibrary(babel.hKJMod); */
  }

  ZeroMemory(&babel, sizeof(babel));

  return 0;
}

static int OhbInitHomeDir(void)
{
  HKEY hKey;
  int len, i;
  DWORD dwLen = OHB_HOME_DIR_LEN;

  if (RegOpenKeyExA(HKEY_LOCAL_MACHINE,
    "Software\\Unisoft\\Babel TOP 2002\\JSender", 0, KEY_QUERY_VALUE, &hKey)
    != ERROR_SUCCESS)
  {
    return -1;
  }

  if (RegQueryValueExA(hKey, NULL, NULL, NULL, (LPBYTE)babel.szJKHomeDir,
    &dwLen) != ERROR_SUCCESS)
  {
    return -2;
  }

  RegCloseKey(hKey);

  babel.szJKHomeDir[OHB_HOME_DIR_LEN - 1] = '\0';

  len = lstrlenA(babel.szJKHomeDir);
  if (len > 2000)
  {
    return -3;
  }

  for (i = len - 1; i >= 0; i--)
  {
    if (babel.szJKHomeDir[i] == '\\')
    {
      babel.szJKHomeDir[i] = '\0';
      lstrcatA(babel.szJKHomeDir, "\\dicJK\\");
      lstrcpy(babel.szKJHomeDir, babel.szJKHomeDir);
      babel.szKJHomeDir[i] = '\0';
      lstrcatA(babel.szKJHomeDir, "\\dicKJ\\");
      break;
    }
  }

  if (i < 0)
  {
    return -4;
  }

  return 0;
}

static int OhbDestroyHomeDir(void)
{
  babel.szJKHomeDir[0] = '\0';
  babel.szKJHomeDir[0] = '\0';

  return 0;
}

static LRESULT CALLBACK OhbTransWindowMsgProc(HWND hwnd, UINT message,
  WPARAM wParam, LPARAM lParam)
{
  if (hwnd != NULL &&
    (message == OHB_JK_TRANS_MSG || message == OHB_KJ_TRANS_MSG))
  {
    PostMessage(hwnd, WM_QUIT, 0, 0);
  }

  return DefWindowProc(hwnd, message, wParam, lParam);
}

int OhbBabelJKInit(const OHBSETTING *pSetting)
{
  static int ret;

  if (bInit > 0)
  {
    ++bInit;
    return ret;
  }

  ++bInit;

  return (ret = OhbBabelInit());
}

void OhbBabelJKDestroy(void)
{
  --bInit;

  if (bInit > 0)
  {
    return;
  }

  OhbBabelDestroy();
}

int OhbBabelKJInit(const OHBSETTING *pSetting)
{
  return OhbBabelJKInit(pSetting);
}

void OhbBabelKJDestroy(void)
{
  OhbBabelJKDestroy();
}

const OHBTRANS *OhbBabelJKTrans(LPCWSTR jpStr, const OHBSETTING *pSetting)
{
  LPSTR szSrc;
  LPWSTR szDst;
  LPCSTR szTrans;
  int len;
  MSG msg;
  int fd;
  HWND hwndTransWindow;
  OHBTRANS *ohbTrans;

  OHBLOCK();

  len = WideCharToMultiByte(932, 0, jpStr, -1, 0, 0, 0, 0);
  szSrc = HeapAlloc(GetProcessHeap(), 0, len);
  WideCharToMultiByte(932, 0, jpStr, -1, szSrc, len, 0, 0);

  hwndTransWindow = OhbCreateTransWindow();
  fd = babel.InitJKTE(hwndTransWindow, babel.szJKHomeDir, 0);
  babel.SetJKTESource(fd, szSrc);
  babel.SetJKTEOption(fd, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, szSrc);
  babel.TransJKTE(fd, 0);
  while (GetMessage(&msg, hwndTransWindow, 0, 0))
  {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }
  OhbDestroyTransWindow(hwndTransWindow);
  szTrans = babel.GetJKTESolution(fd);

  len = MultiByteToWideChar(949, 0, szTrans, -1, 0, 0);
  szDst = HeapAlloc(GetProcessHeap(), 0, len * 2);
  MultiByteToWideChar(949, 0, szTrans, -1, szDst, len);

  ohbTrans = HeapAlloc(GetProcessHeap(), 0, sizeof(OHBTRANS));
  ohbTrans->pSetting = pSetting;
  ohbTrans->pData = (LPVOID)fd;
  ohbTrans->lpStr = szDst;
  HeapFree(GetProcessHeap(), 0, szSrc);

  OHBUNLOCK();

  return ohbTrans;
}

void OhbBabelJKFree(const OHBTRANS *ohbTrans)
{
  int fd;

  OHBLOCK();
  fd = (int)ohbTrans->pData;
  babel.ReleaseJKTESolution(fd);
  babel.EndJKTE(fd, fd);
  HeapFree(GetProcessHeap(), 0, (LPVOID)ohbTrans->lpStr);
  HeapFree(GetProcessHeap(), 0, (LPVOID)ohbTrans);
  OHBUNLOCK();
}

const OHBTRANS *OhbBabelKJTrans(LPCWSTR krStr, const OHBSETTING *pSetting)
{
  LPSTR szSrc;
  LPWSTR szDst;
  LPCSTR szTrans;
  int len;
  MSG msg;
  int fd;
  HWND hwndTransWindow;
  OHBTRANS *ohbTrans;

  OHBLOCK();

  len = WideCharToMultiByte(949, 0, krStr, -1, 0, 0, 0, 0);
  szSrc = HeapAlloc(GetProcessHeap(), 0, len);
  WideCharToMultiByte(949, 0, krStr, -1, szSrc, len, 0, 0);

  hwndTransWindow = OhbCreateTransWindow();
  fd = babel.InitKJTE(hwndTransWindow, babel.szKJHomeDir, 0);
  babel.SetKJTESource(fd, szSrc);
  babel.SetKJTEOption(fd, 0, 0, 0, 0, 0, szSrc);
  babel.TransKJTE(fd, 0);
  while (GetMessage(&msg, hwndTransWindow, 0, 0))
  {
    TranslateMessage(&msg);
    DispatchMessage(&msg);
  }
  OhbDestroyTransWindow(hwndTransWindow);
  szTrans = babel.GetKJTESolution(fd);

  len = MultiByteToWideChar(20932, 0, szTrans, -1, 0, 0);
  szDst = HeapAlloc(GetProcessHeap(), 0, len * 2);
  MultiByteToWideChar(20932, 0, szTrans, -1, szDst, len);

  ohbTrans = HeapAlloc(GetProcessHeap(), 0, sizeof(OHBTRANS));
  ohbTrans->pSetting = pSetting;
  ohbTrans->pData = (LPVOID)fd;
  ohbTrans->lpStr = szDst;
  HeapFree(GetProcessHeap(), 0, szSrc);

  OHBUNLOCK();

  return ohbTrans;
}

void OhbBabelKJFree(const OHBTRANS *ohbTrans)
{
  int fd;

  OHBLOCK();
  fd = (int)ohbTrans->pData;
  babel.ReleaseKJTESolution(fd);
  babel.EndKJTE(fd, fd);
  HeapFree(GetProcessHeap(), 0, (LPVOID)ohbTrans->lpStr);
  HeapFree(GetProcessHeap(), 0, (LPVOID)ohbTrans);
  OHBUNLOCK();
}
